/*
    This file is part of Wordle Solver.
    Copyright (C) 2022  Laurent Poirrier

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include "word.h"
#include "dict.h"

// ****************************************************************************
// 
// ****************************************************************************
void dict_init(dict *d)
{
	d->count = 0;
	d->alloc = 0;
	d->words = NULL;
}

void dict_clear(dict *d)
{
	free(d->words);
}

// ****************************************************************************
// 
// ****************************************************************************
void dict_grow_nt(dict *d, size_t xc)
{
	size_t da = d->alloc;
	
	if (da < DICT_ALLOC_MIN)
		da = DICT_ALLOC_MIN;
	
	while (xc > da) {
		size_t da1 = da * 2;
		
		if (da1 <= da) {
			fprintf(stderr, "realloc failed "
				"(%zd -> %zd > %zd words)\n",
				d->alloc, xc, da);
			exit(1);
		}
		
		da = da1;
	}
	
	word *t = realloc(d->words, da * sizeof(word));
	
	if (!t) {
		fprintf(stderr, "realloc failed (%zd -> %zd words)\n",
			d->alloc, da);
		exit(1);
	}
	
	d->words = t;
	d->alloc = da;
}

// ****************************************************************************
// 
// ****************************************************************************
void dict_append_dict(dict *dst, const dict *src)
{
	size_t dc = dst->count;
	size_t sc = src->count;
	size_t xc = dst->count + src->count;
	
	dict_grow(dst, xc);
	
	word *dw = dst->words;
	word *sw = src->words;
	
	for (size_t i = 0; i < sc; i++) {
		dw[dc + i] = sw[i];
	}
	
	dst->count = xc;
}

// ****************************************************************************
// 
// ****************************************************************************
void dict_append_file(dict *d, const char *path)
{
	FILE *f = fopen(path, "r");
	
	if (!f) {
		fprintf(stderr, "%s: %s\n", path, strerror(errno));
		exit(1);
	}
	
	size_t count = 0;
	size_t dropped = 0;
	char buff[LENGTH + 3];
	buff[LENGTH + 2] = 0;
	
	while (1) {
		char *s = fgets(buff, LENGTH + 2, f);
		
		if (!s)
			break;
		
		size_t l = strlen(s);
		
		if ((l > 0) && (s[l - 1] == '\n')) {
			s[l - 1] = 0;
			l--;
		}
		
		if (l != LENGTH) {
			dropped++;
			continue;
		}
		
		dict_append_word(d, word_build(s));
		count++;
	}
	
	fprintf(stderr, "%s: %d-words: %zd, dropped: %zd\n",
		path, LENGTH, count, dropped);
}

// ****************************************************************************
// 
// ****************************************************************************
void dict_lower(dict *d)
{
	size_t dc = d->count;
	
	word *dw = d->words;
	
	for (size_t i = 0; i < dc; i++) {
		dw[i] = word_lower(dw[i]);
	}
}

// ****************************************************************************
// 
// ****************************************************************************
static int dict_cmp(const void *a, const void *b)
{
	return __builtin_memcmp(a, b, LENGTH);
}

void dict_sort(dict *d)
{
	qsort(d->words, d->count, sizeof(word), dict_cmp);
}

// ****************************************************************************
// 
// ****************************************************************************
void dict_filter_dupes(dict *dst, const dict *src)
{
	if (dst == src) {
		size_t dc = dst->count;
		size_t xc = 1;
		
		word *dw = dst->words;
		word cur = dw[0];
		
		for (size_t i = 1; i < dc; i++) {
			if (!word_is(dw[i], cur)) {
				dw[xc] = dw[i];
				cur = dw[xc];
				xc++;
			}
		}
		
		dst->count = xc;
		
		return;
	}

	dict_reset(dst);
	size_t sc = src->count;
	word *sw = src->words;
	word cur = sw[0];
	
	for (size_t i = 0; i < sc; i++) {
		word w = sw[i];
		if (!word_is(w, cur)) {
			dict_append_word(dst, w);
			cur = w;
		}
	}
}

// ****************************************************************************
// 
// ****************************************************************************
void dict_filter_letters(dict *dst, const dict *src)
{
	if (dst == src) {
		size_t dc = dst->count;
		size_t xc = 0;
		
		word *dw = dst->words;
		
		for (size_t i = 0; i < dc; i++) {
			if (word_letters(dw[i])) {
				dw[xc] = dw[i];
				xc++;
			}
		}
		
		dst->count = xc;
		
		return;
	}

	dict_reset(dst);
	size_t sc = src->count;
	word *sw = src->words;
	
	for (size_t i = 0; i < sc; i++) {
		if (word_letters(sw[i])) {
			dict_append_word(dst, sw[i]);
		}
	}
}

// ****************************************************************************
// 
// ****************************************************************************
void dict_filter_pattern(dict *dst, const dict *src, word fix, lset contain)
{
	if (dst == src) {
		size_t dc = dst->count;
		size_t xc = 0;
	
		word *dw = dst->words;
	
		for (size_t i = 0; i < dc; i++) {
			if ((word_fix_match(dw[i], fix))
			&&  (word_has_all(dw[i], contain))) {
				dw[xc] = dw[i];
				xc++;
			}
		}
		
		dst->count = xc;
		
		return;
	}
	
	dict_reset(dst);
	size_t sc = src->count;
	word *sw = src->words;
	
	for (size_t i = 0; i < sc; i++) {
		if ((word_fix_match(sw[i], fix))
		&&  (word_has_all(sw[i], contain))) {
			dict_append_word(dst, sw[i]);
		}
	}
}

// ****************************************************************************
// 
// ****************************************************************************
// this is quadratic but good enough for now
void dict_subtract(dict *dst, const dict *a, const dict *b)
{
	if (dst == a) {
		size_t dc = dst->count;
		size_t bc = b->count;
		size_t xc = 0;
		
		word *dw = dst->words;
		word *bw = b->words;
		
		for (size_t i = 0; i < dc; i++) {
			word dwi = dw[i];
			int found = 0;
			
			for (size_t j = 0; j < bc; j++) {
				if (word_is(dwi, bw[j])) {
					found = 1;
					//printf("- \"%s\",\n",
					//	word_str(dwi).s);
					//fflush(stdout);
					break;
				}
			}
			
			if (!found) {
				//printf("+ \"%s\",\n",
				//	word_str(dwi).s);
				//fflush(stdout);
				dw[xc] = dwi;
				xc++;
			}
		}
		
		dst->count = xc;
		
		return;
	}
	
	dict_reset(dst);
	
	size_t ac = a->count;
	size_t bc = b->count;
	
	word *aw = a->words;
	word *bw = b->words;
	
	for (size_t i = 0; i < ac; i++) {
		word awi = aw[i];
		int found = 0;
		
		for (size_t j = 0; j < bc; j++) {
			if (word_is(awi, bw[j])) {
				found = 1;
				break;
			}
		}
		
		if (!found) {
			dict_append_word(dst, awi);
		}
	}
}

// ****************************************************************************
// 
// ****************************************************************************
dict_index dict_index_get(const dict *d, const word w)
{
	size_t dc = d->count;
	word *dw = d->words;
	
	for (dict_index i = 0; i < dc; i++) {
		if (word_is(dw[i], w))
			return i;
	}
	
	return DICT_NONE;
}

int dict_has(const dict *d, const word w)
{
	return (dict_index_get(d, w) != DICT_NONE);
}

// ****************************************************************************
// 
// ****************************************************************************
