/*
    This file is part of Wordle Solver.
    Copyright (C) 2022  Laurent Poirrier

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include "check.h"
#include "dict.h"
#include "lookup.h"

// ****************************************************************************
// 
// ****************************************************************************
/*
  The rule for when there are duplicate letters in guess is as follows:

  1. Green <=> matching letters, that part is easy.

  2. Then, the total green+yellow count of a letter in the guess must match
     the number of occurences of the letter in the target.

  Example: target = ROBOT

     POTOO
  -> .GYG. (boths O's are green - easy)

     OVOLO
  -> Y.Y.. (first two of the 3 O's are yellow)

     ERROR
  -> .Y.G. (one of the 3 R's is yellow)

     LOOPY
  -> .GY.. (on O is green, the other is yellow)

  --------------------

  Note that hard mode is aware of this. After guessing LOOPY (getting a
  green O and a yellow O), we cannot play FORGE (just a green O).

  Check:
    LOOPY
    .GY..

    MOTOR
    .GYGY

    TOROS
    YGYG.

    ROBOT
    GGGGG

  Also:
    OVOLO
    Y.Y..

  now we need two O's !!!
*/
static hash_index lut_play_hash(word guess, word target)
{
	// information increase
	word color = word_zero();
	
	for (size_t j = 0; j < LENGTH; j++) {
		if (guess.s[j] == target.s[j]) {
			color.s[j] = 2;
			target.s[j] = 0;
			guess.s[j] = 0;
		}
	}
	
	for (size_t j = 0; j < LENGTH; j++) {
		uint8_t l = guess.s[j];
		
		if (l == 0)
			continue;
		
		uint8_t c = 0;

		for (size_t k = 0; k < LENGTH; k++) {
			if (target.s[k] == l) {
				target.s[k] = 0;
				c = 1;
				break;
			}
		}
		
		color.s[j] = c;
	}
	
	hash_index hash = 0;
	for (size_t j = 0; j < LENGTH; j++) {
		hash = (hash * 3) + color.s[j];
	}
	
	return hash;
}

// ****************************************************************************
// 
// ****************************************************************************
word lut_hash_word(hash_index h)
{
	if (h == HASH_NONE)
		return word_build("/////");
	if (h == HASH_ANY)
		return word_build("*****");
	
	word a = word_zero();
	
	for (size_t k = 0; k < LENGTH; k++) {
		size_t j = (LENGTH - 1) - k;
		
		uint8_t c = h % 3;
		h = h / 3;

		a.s[j] = ((c == 0) ? '.' : ((c == 1) ? 'y' : 'G'));
	}
	
	return a;
}


// ****************************************************************************
// 
// ****************************************************************************
void lut_build(lut *t, dict *d)
{
	// checks
	if (d->count != DICT_SIZE) {
		fprintf(stderr,
			"lookup dict has size %zd != %d (header defined)\n",
			d->count, DICT_SIZE);
		exit(1);
	}
	
	// build table of hashes
	word *w = d->words;
	hash_index (*hash)[DICT_SIZE] = t->hash;
	dict_index answer[THREE_TO_LENGTH];
	hash_index disamb_max = 0;
	
	for (dict_index i = 0; i < DICT_SIZE; i++) {
		word guess = w[i];
		
		for (hash_index a = 0; a < THREE_TO_LENGTH; a++)
			answer[a] = 0;
		
		for (dict_index j = 0; j < DICT_SIZE; j++) {
			word target = w[j];
			
			hash_index h = lut_play_hash(guess, target);
			hash[i][j] = h;
			
			answer[h]++;
		}
		
		hash_index disamb = 0;
		
		for (hash_index a = 0; a < THREE_TO_LENGTH; a++)
			disamb += (answer[a] != 0);
		
		//printf("%s: disambiguation %d\n", word_str(w[i]).s, disamb);
		if (disamb > disamb_max) {
			disamb_max = disamb;
		}
	}
	
	check(disamb_max, ==, DISAMBIGUATION);
	
	//printf("max disambiguation: %d\n", disamb_max);

	// build table of hash words
	for (hash_index h = 0; h < THREE_TO_LENGTH; h++) {
		t->hash_word[h] = lut_hash_word(h);
	}
	
	// build table of masks
	for (dict_index i = 0; i < DICT_SIZE; i++) {
		word w = d->words[i];
		lset all = LSET_ZERO;
		
		for (size_t l = 0; l < LENGTH; l++) {
			lset one = lset_one(w.s[l]);
			t->word_mask[i].fix[l] = one;
			all |= one;
		}

		t->word_mask[i].contain = all;
	}
	
	// build hash inversion table
	for (hash_index h = 0; h < THREE_TO_LENGTH; h++) {
		hash_index hash = h;
		for (size_t k = 0; k < LENGTH; k++) {
			size_t j = (LENGTH - 1) - k;
			hash_index a = hash % 3;
			hash = hash / 3;
			
			t->fix[h][j] = (a == 2) ? LSET_ZERO : LSET_ALL;
			t->contain[h][j] = (a != 0) ? LSET_ALL : LSET_ZERO;
		}
	}
}


// ****************************************************************************
// 
// ****************************************************************************
hash_index lut_hash_parse(word a)
{
	hash_index hash = 0;
	
	for (size_t j = 0; j < LENGTH; j++) {
		uint8_t l = a.s[j];
		uint8_t c = ((l == 'G') || (l == 'g')) ? 2
			: (((l == 'Y') || (l == 'y')) ? 1 : 0);
		
		hash = (hash * 3) + c;
	}
	
	return hash;
}


// ****************************************************************************
// 
// ****************************************************************************
void lut_print_hash(FILE *f, hash_index h)
{
	fprintf(f, "hash %d: ", h);
	
	uint8_t r[LENGTH];
	
	for (size_t k = 0; k < LENGTH; k++) {
		size_t j = (LENGTH - 1) - k;
		
		r[j] = h % 3;
		h = h / 3;
	}
	
	for (size_t j = 0; j < LENGTH; j++) {
		fprintf(f, "%c",
			(r[j] == 2) ? '#' : (r[j] == 1) ? '^' : '.');
	}
	fprintf(f, "\n");
}


// ****************************************************************************
// 
// ****************************************************************************
/*
static void lset_print(FILE *f, lset s)
{
	for (uint8_t l = 'a'; l <= 'z'; l++) {
		fprintf(f, "%c",
			(s & lset_one(l)) ? l : '.');
	}
}

void hard_print(FILE *f, mask m)
{
	for (size_t l = 0; l < LENGTH; l++) {
		fprintf(f, "fix %3zd: ", l);
		lset_print(f, m.fix[l]);
		fprintf(f, "\n");
	}
	fprintf(f, "contain: ");
	lset_print(f, m.contain);
	fprintf(f, "\n");
}
*/
// ****************************************************************************
// 
// ****************************************************************************
int hard_allowed(lut *t,
	word prev, hash_index prev_hash,
	word guess)
{
	word a = t->hash_word[prev_hash];

	for (size_t j = 0; j < LENGTH; j++) {
		if (a.s[j] == 'G') {
			if (guess.s[j] != prev.s[j]) {
				//printf("G fail in pos %zd %c\n",
				//	j, prev.s[j]);
				return 0;
			}
			prev.s[j] = 0;
			guess.s[j] = 0;
		}
	}
	
	for (size_t j = 0; j < LENGTH; j++) {
		if (a.s[j] == 'y') {
			uint8_t c = prev.s[j];
			
			for (size_t k = 0; ; k++) {
				if (k >= LENGTH) {
					//printf("Y fail pos %zd %c\n",
					//	j, c);
					return 0;
				}
				
				if (guess.s[k] == c) {
					guess.s[k] = 0;
					break;
				}
			}
		}
	}
	
	return 1;
}

// ****************************************************************************
// 
// ****************************************************************************

