/*
    This file is part of Wordle Solver.
    Copyright (C) 2022  Laurent Poirrier

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/
#define _DEFAULT_SOURCE
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "check.h"
#include "dict.h"
#include "builtin_dict.h"
#include "play.h"
#include "solve.h"
#include "tree.h"
#include "main.h"

// ****************************************************************************
// 
// ****************************************************************************
#define HELP_MSG \
"Usage: wordle {play|solve|build|check} [options]\n" \
"\n" \
"   play        play one Wordle game\n" \
"  solve        solve one Wordle game\n" \
"  build        build Wordle decision tree\n" \
"  check        play all possible Wordle games to check optimal tree\n" \
"\n" \
"Options:\n" \
"\n" \
" -hard         hard mode\n" \
" -cheat        cheat mode\n" \
" -w word       target word to play\n" \
" -t tid tnum   parallel enumeration thread\n" \
" -d depth      enumeration depth\n" \
" -k K          heuristic search, K candidates\n" \
" -c C          heuristic search, |X| <= C for disambiguation\n" \
" -i nodes      progress display interval\n" \
"\n" \
"\n"


// ****************************************************************************
// Main
// ****************************************************************************
#define CMD_NONE	0
#define CMD_PLAY	1
#define CMD_SOLVE	2
#define CMD_BUILD	3
#define CMD_CHECK	4
#define CMD_EXPORT_JS	5
#define CMD_EXPORT_TXT	6
int main(int argc, char **argv)
{
	int cmd = -1;
	int flags = 0;
	size_t tid = 0;
	size_t tnum = 1;
	depth_index depth = 6;
	dict_index fast_k = 0;
	dict_index fast_c = 0;
	node_index prog_i = 0;
	char *target = NULL;
	
	for (int t = 1; t < argc; t++) {
		char *a = argv[t];
		
		if (!strcmp(a, "play")) {
			cmd = CMD_PLAY;
		} else if (!strcmp(a, "solve")) {
			cmd = CMD_SOLVE;
		} else if (!strcmp(a, "build")) {
			cmd = CMD_BUILD;
		} else if (!strcmp(a, "check")) {
			cmd = CMD_CHECK;
		} else if (!strcmp(a, "export-js")) {
			cmd = CMD_EXPORT_JS;
		} else if (!strcmp(a, "export-txt")) {
			cmd = CMD_EXPORT_TXT;
		} else if (!strcmp(a, "-hard")) {
			flags |= GAME_HARD;
		} else if (!strcmp(a, "-cheat")) {
			flags |= GAME_CHEAT;
		} else if ((t + 1 < argc) && (!strcmp(a, "-w"))) {
			target = argv[++t];
		} else if ((t + 2 < argc) && (!strcmp(a, "-t"))) {
			tid = strtol(argv[++t], NULL, 0);
			tnum = strtol(argv[++t], NULL, 0);
		} else if ((t + 1 < argc) && (!strcmp(a, "-d"))) {
			depth = strtol(argv[++t], NULL, 0);
		} else if ((t + 1 < argc) && (!strcmp(a, "-k"))) {
			fast_k = strtol(argv[++t], NULL, 0);
		} else if ((t + 1 < argc) && (!strcmp(a, "-c"))) {
			fast_c = strtol(argv[++t], NULL, 0);
		} else if ((t + 1 < argc) && (!strcmp(a, "-i"))) {
			prog_i = strtol(argv[++t], NULL, 0);
		} else {
			printf("%s", HELP_MSG);
			return 1;
		}
		
	}
	
	
	if (cmd == CMD_PLAY) {
		game g;
		
		game_init(&g, flags);
		
		if (target) {
			game_set_word(&g, target);
		} else {
			srandom(time(NULL));
			game_set_random(&g);
		}
		
		game_interactive(&g);
		
		game_clear(&g);
	} else if (cmd == CMD_SOLVE) {
		solver s;
		
		solver_init(&s, flags);
		
		solver_interactive(&s);
		
		solver_clear(&s);
	} else if (cmd == CMD_BUILD) {
		time_t start = time(NULL);
		
		tree *t = tree_build(tid, tnum, depth,
			fast_k, fast_c, prog_i, flags);
		
		time_t stop = time(NULL);
		
		tree_free(t);
		
		fprintf(stderr, "Elapsed: %zd seconds\n",
			(size_t)(stop - start));
	} else if (cmd == CMD_CHECK) {
		dict d;
		
		dict_init(&d);
		dict_builtin_full(&d);
		
		size_t count0 = d.count;
		printf("      Raw: %zd    (builtin)\n", d.count);
		dict_lower(&d);
		dict_filter_dupes(&d, &d);
		printf(" No-dupes: %zd\n", d.count);
		dict_filter_letters(&d, &d);
		printf("  Letters: %zd\n", d.count);
		
		check(d.count, ==, count0);

		solver s;
		
		solver_init(&s, flags);
		solver_all(&s);
		solver_connectivity(&s);
		solver_clear(&s);
	} else if (cmd == CMD_EXPORT_JS) {
		solver_export_js(flags);
	} else if (cmd == CMD_EXPORT_TXT) {
		solver_export_txt(flags);
	} else {
		printf("%s", HELP_MSG);
		return 1;
	}
	
	return 0;
}

