/*
    This file is part of Wordle Solver.
    Copyright (C) 2022  Laurent Poirrier

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/
#define _DEFAULT_SOURCE
#include <string.h>
#include <stdlib.h>
#include "check.h"
#include "dict.h"
#include "builtin_dict.h"
#include "lookup.h"
#include "play.h"

// ****************************************************************************
// 
// ****************************************************************************
typedef struct shortstr_t {
	char s[256];
} shortstr;


static shortstr colorize(char *str, char *color)
{
	shortstr r;
	size_t i = 0;
	size_t o = 0;
	int active = 1;
	
	while (1) {
		char t = str[i];
		char c = (active) ? color[i] : ' ';
		
		i++;
		
		if (t == 0) {
			r.s[o] = 0;
			return r;
		}
		
		if (c == 0)
			active = 0;
		
		if (o + 10 >= sizeof(r.s)) {
			r.s[o] = 0;
			return r;
		}
			
		if ((c == 'G') || (c == 'g')) {
			r.s[o + 0] = 27;
			r.s[o + 1] = '[';
			r.s[o + 2] = '9';
			r.s[o + 3] = '2';
			r.s[o + 4] = 'm';
			r.s[o + 5] = t;
			r.s[o + 6] = 27;
			r.s[o + 7] = '[';
			r.s[o + 8] = '0';
			r.s[o + 9] = 'm';
			o += 10;
			continue;
		}
		
		if ((c == 'Y') || (c == 'y')) {
			r.s[o + 0] = 27;
			r.s[o + 1] = '[';
			r.s[o + 2] = '9';
			r.s[o + 3] = '3';
			r.s[o + 4] = 'm';
			r.s[o + 5] = t;
			r.s[o + 6] = 27;
			r.s[o + 7] = '[';
			r.s[o + 8] = '0';
			r.s[o + 9] = 'm';
			o += 10;
			continue;
		}
		
		if (c != ' ') {
			r.s[o + 0] = 27;
			r.s[o + 1] = '[';
			r.s[o + 2] = '3';
			r.s[o + 3] = '1';
			r.s[o + 4] = 'm';
			r.s[o + 5] = t;
			r.s[o + 6] = 27;
			r.s[o + 7] = '[';
			r.s[o + 8] = '0';
			r.s[o + 9] = 'm';
			o += 10;
			continue;
		}

		r.s[o] = t;
		o++;
	}
}

// ****************************************************************************
// 
// ****************************************************************************
void game_init(game *g, int flags)
{
	g->flags = flags;
	
	dict_init(&g->allow);
	dict_builtin_full(&g->allow);
	
	g->table = malloc(sizeof(lut));
	if (!g->table) {
		fprintf(stderr, "malloc(table) failed\n");
		exit(1);
	}
	
	lut_build(g->table, &g->allow);
	
	g->target = word_zero();
	g->target_index = DICT_NONE;
	g->round = 0;
	g->prev_g = word_zero();
	g->prev_h = HASH_NONE;
	g->green = g->yellow = g->gray = 0;
	g->hist_l = 0;
}

void game_reset(game *g)
{
	g->target = word_zero();
	g->target_index = DICT_NONE;
	g->round = 0;
	g->prev_g = word_zero();
	g->prev_h = HASH_NONE;
	g->green = g->yellow = g->gray = 0;
	g->hist_l = 0;
}

void game_clear(game *g)
{
	free(g->table);
	dict_clear(&g->allow);
}

// ****************************************************************************
// 
// ****************************************************************************
void game_set_word(game *g, const char *s)
{
	game_reset(g);
	
	size_t l = strlen(s);
	
	if (l != LENGTH) {
		fprintf(stderr, "play_set_word(): '%s' not of length %d\n",
			s, LENGTH);
		exit(1);
	}
	
	g->target = word_build(s);
	g->target_index = dict_index_get(&g->allow, g->target);
	
	if (g->target_index == DICT_NONE) {
		fprintf(stderr, "play_set_word(): '%s' not an allowed word\n",
			s);
		exit(1);
	}
}

// ****************************************************************************
// 
// ****************************************************************************
void game_set_random(game *g)
{
	game_reset(g);
	
	if (g->flags & GAME_CHEAT) {
		dict sol;
		
		dict_init(&sol);
		dict_builtin_solutions(&sol);
		
		size_t i = ((size_t)random()) % sol.count;
		
		g->target = sol.words[i];
		g->target_index = dict_index_get(&g->allow, g->target);
		
		if (g->target_index == DICT_NONE) {
			fprintf(stderr, "play_set_random(): "
				"'%s' not an allowed word\n",
				word_str(g->target).s);
			exit(1);
		}

		dict_clear(&sol);
	} else {
		size_t i = ((size_t)random()) % g->allow.count;
		
		g->target_index = i;
		g->target = g->allow.words[i];
	}
}

// ****************************************************************************
// 
// ****************************************************************************
void game_keyboard(game *g)
{
	char *row[3] = {
		"q  w  e  r  t  y  u  i  o  p",
		" a  s  d  f  g  h  j  k  l",
		"  z  x  c  v  b  n  m"
	};
	
	for (int t = 0; t < 3; t++) {
		char *r = row[t];
		
		printf(" ");
		
		while (*r) {
			char t = *r;
			r++;

			if (t == ' ') {
				printf(" ");
				continue;
			}
			
			lset letter = lset_one(t);
			
			char c = 0;
			
			if (g->green & letter)
				c = 'G';
			else if (g->yellow & letter)
				c = 'y';
			else if (g->gray & letter)
				c = '.';
			else
				c = ' ';
			
			char str[2] = { t, 0 };
			char color[2] = { c, 0 };

			printf("%s", colorize(str, color).s);
		}
		
		printf("\n");
	}
}
// ****************************************************************************
// Play
// ****************************************************************************
void game_interactive(game *g)
{
	int win = 0;
	
	while (1) {
		printf("\n");

		for (int t = 0; t < g->hist_l; t++) {
			word gl = g->allow.words[g->hist_g[t]];
			word hl = g->table->hash_word[g->hist_h[t]];
			
			printf("  %s\n",
				colorize(word_str(gl).s, word_str(hl).s).s);
		}
		
		printf("\n");
		
		if (win) {
			printf("    Win!\n\n");
			break;
		}
		
		game_keyboard(g);
		
		if (g->hist_l >= PLAY_MAX) {
			printf("    Max %d guesses\n", PLAY_MAX);
			break;
		}
		
		printf("\n> ");
		fflush(stdout);
		
		char buff[LENGTH + 3];
		buff[LENGTH + 2] = 0;
		
		char *s = fgets(buff, LENGTH + 2, stdin);
		
		if (!s) {
			printf("    Stop\n\n");
			break;
		}
		
		size_t l = strlen(s);
		
		if ((l > 0) && (s[l - 1] == '\n'))
			s[--l] = 0;
		
		if ((strcmp(s, "r") == 0) || (strcmp(s, "R") == 0)) {
			printf("\n    *** restart ***\n\n");
			word t = g->target;
			dict_index ti = g->target_index;
			game_reset(g);
			g->target = t;
			g->target_index = ti;
			continue;
		}
		
		if ((strcmp(s, "q") == 0) || (strcmp(s, "Q") == 0)) {
			printf("    Stop\n\n");
			break;
		}

		if (l != LENGTH) {
			printf("    Not %d letters: '%s'\n",
				LENGTH, s);
			continue;
		}
		
		word guess = word_lower(word_build(s));
		
		dict_index i = dict_index_get(&g->allow, guess);
		
		if (i == DICT_NONE) {
			printf("    Not a known word\n");
			continue;
		}
		
		if ((g->flags & GAME_HARD) && (g->round > 0)) {
			if (!hard_allowed(g->table,
					g->prev_g, g->prev_h, guess)) {
				printf("    Does not follow hard mode hints\n");
				continue;
			}
		}
		
		hash_index h = lut_lookup(g->table, i, g->target_index);
		word answer = g->table->hash_word[h];
		
		g->hist_g[g->hist_l] = i;
		g->hist_h[g->hist_l] = h;
		g->hist_l++;
		
		if (i == g->target_index)
			win = 1;
		
		g->prev_g = guess;
		g->prev_h = h;

		for (size_t j = 0; j < LENGTH; j++) {
			if (answer.s[j] == 'G') {
				g->green |= lset_one(guess.s[j]);
			} else if (answer.s[j] == 'y') {
				g->yellow |= lset_one(guess.s[j]);
			} else {
				g->gray |= lset_one(guess.s[j]);
			}
		}
		
		g->round++;
	}
}

