/*
    This file is part of Wordle Solver.
    Copyright (C) 2022  Laurent Poirrier

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/
#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include "check.h"
#include "aux_sort.h"
#include "dict.h"
#include "builtin_dict.h"
#include "lookup.h"
#include "tree.h"

// ****************************************************************************
// 
// ****************************************************************************
#define NODE_MAX		((node_index)8 << 20)

#define TREE_STORE

// ****************************************************************************
// 
// ****************************************************************************
/*
  Local variables of tree_dfs() are stored in an explicit stack
  for two reasons:
  
  1. It allows the garbage collector to update node pointers
     (cur_node and best_node).
  
  2. It allows tree_progress() to inspect the state of all the
     recursion levels.
*/
typedef struct local_t {
	depth_index depth_ub, depth_max, best_depth;
	hash_index answer_count, a, h;
	dict_index couldbe_size, guesses_size;
	dict_index guesses_valid;
	dict_index ord_cutoff, ord, g, asize;
	node_index cur_node, best_node;
} local;

typedef struct level_t {
	local v;
	
	hash_index answer_list[THREE_TO_LENGTH];

	hash_index hash_cache[DICT_SIZE][DICT_SIZE];
	uint64_t guess_score[DICT_SIZE];
	dict_index guesses_sorted[DICT_SIZE];
	dict_index guesses_aux[DICT_SIZE];
	
	dict_index answer_couldbe_size[THREE_TO_LENGTH];
	dict_index answer_couldbe_offs[THREE_TO_LENGTH];
	dict_index answer_couldbe_set[DICT_SIZE];
} level;

// >1 GB, do not attempt to allocate on stack
typedef struct tree_t {
	int flags;
	time_t start_time, prog_time;
	dict_index fast_k, fast_c;
	size_t prog_i, prog_ops, prog_last, prog_next;
	
	dict dictionary;
	dict_index couldbe[DICT_SIZE];
	
	depth_index global_depth_ub;
	node_index root;
	node_index node_top, node_active, node_total;
	size_t thread_id, thread_total;
	
	lut table;
	node node_buffer[NODE_MAX];
	node_index remap[NODE_MAX];
	level stack[];
} tree;

// ****************************************************************************
// 
// ****************************************************************************
static void tree_progress(tree *t, depth_index depth, const char *note)
{
	if (t->prog_ops < t->prog_next)
		return;
	
	time_t cur_time = time(NULL);
	time_t g_delta = cur_time - t->start_time;
	time_t p_delta = cur_time - t->prog_time;
	size_t ops_delta = t->prog_ops - t->prog_last;
	
	if (g_delta < 1)
		g_delta = 1;
	if (p_delta < 1)
		p_delta = 1;
	
	fprintf(stderr, "#  Elapsed: %5zds, op: %6zd G, "
		"op/s: %5zd M current | %5zd M global\n",
		g_delta, t->prog_ops >> 30,
		(ops_delta / p_delta) >> 20,
		(t->prog_ops / g_delta) >> 20);

	t->prog_time = cur_time;
	t->prog_last = t->prog_ops;
	
	t->prog_next += t->prog_i;
	if (t->prog_next < t->prog_ops)
		t->prog_next = t->prog_ops + t->prog_i;
	
	
	fprintf(stderr, "node %12zd | in-tree %7zd / in-buffer %7zd\n",
		t->node_total, t->node_active, t->node_top);
	
	for (depth_index d = 0; d <= depth; d++) {
		local v = t->stack[d].v;

		char bx = ' ';
		char wx = ' ';
		depth_index bd = v.best_depth;
		depth_index wd = v.depth_max;
		
		if (bd == DEPTH_INF) {
			bx = '>';
			bd = v.depth_ub;
		}
		
		if (wd == DEPTH_INF) {
			wx = '>';
			wd = bd;
		}
		
		fprintf(stderr, "%*s'-- %5d/%5d '%s' best %c%d "
			"| ans %3d/%3d size %5d lb %c%d%s\n",
			d * 2, "",
			v.ord, v.guesses_valid,
			word_str(t->dictionary.words[v.g]).s,
			bx, bd,
			v.a, v.answer_count,
			v.asize,
			wx, wd,
			(d == depth) ? note : "");
	}

	for (depth_index d = depth + 1; d < t->global_depth_ub; d++) {
		fprintf(stderr, "\n");
	}
}

// ****************************************************************************
// 
// ****************************************************************************
#ifdef TREE_STORE
static node_index tree_push(tree *t, dict_index guess, depth_index depth)
{
	node *nb = t->node_buffer;
	node_index i = t->node_top;
	
	if (i >= NODE_MAX) {
		fprintf(stderr, "Tree full (%zd nodes)\n", i);
		exit(1);
	}
	
	nb[i].hash = HASH_ANY;
	nb[i].depth = depth;
	nb[i].guess = guess;
	nb[i].child = NODE_NONE;
	nb[i].sibling = NODE_NONE;
	
	t->node_top++;
	t->node_active++;
	t->node_total++;
	
	return i;
}

// *********************************************************************
/*
static void tree_dummy(tree *t, dict_index guess)
{
	(void)guess;
	t->node_total++;
}
*/

// *********************************************************************
static void tree_adopt(tree *t, node_index parent, node_index child,
	hash_index hash)
{
	node *nb = t->node_buffer;
	
	nb[child].hash = hash;
	nb[child].sibling = nb[parent].child;
	nb[parent].child = child;
}

// *********************************************************************
static void tree_prune(tree *t, node_index i)
{
	node *nb = t->node_buffer;
	
	check(nb[i].hash, !=, HASH_NONE);
	nb[i].hash = HASH_NONE;
	t->node_active--;
	
	node_index j = nb[i].child;
	
	while (j != NODE_NONE) {
		node_index next = nb[j].sibling;
		
		tree_prune(t, j);
		
		j = next;
	}
}

// *********************************************************************
/*
static void tree_check(tree *t, char *msg)
{
	node *nb = t->node_buffer;
	node_index o = 0;

	for (node_index i = 0; i < t->node_top; i++) {
		if (nb[i].hash != HASH_NONE) {
			o++;
		}
	}
	
	check_s(o, ==, t->node_active, "Fail at %s\n", msg);
}
*/

static void tree_gc(tree *t, int force)
{
	node_index node_top = t->node_top;
	
	if (!force) {
		if (node_top + 1 + (THREE_TO_LENGTH * 2) <= NODE_MAX)
			return;
	}
	
	fprintf(stderr, "Garbage collection: ");
	
	node_index *remap = t->remap;
	node *nb = t->node_buffer;

	node_index o = 0;
	
	for (node_index i = 0; i < node_top; i++) {
		if (nb[i].hash == HASH_NONE) {
			remap[i] = NODE_NONE;
		} else {
			remap[i] = o;
			o++;
		}
	}
	
	check(o, ==, t->node_active);
	
	if (t->root != NODE_NONE)
		t->root = remap[t->root];
	
	for (depth_index d = 0; d < t->global_depth_ub; d++) {
		level *l = &t->stack[d];
		
		if (l->v.cur_node != NODE_NONE)
			l->v.cur_node = remap[l->v.cur_node];
		if (l->v.best_node != NODE_NONE)
			l->v.best_node = remap[l->v.best_node];
	}
	
	o = 0;
	
	for (node_index i = 0; i < node_top; i++) {
		if (nb[i].hash == HASH_NONE)
			continue;
		
		node ni = nb[i];
		
		if (ni.child != NODE_NONE)
			ni.child = remap[ni.child];
		if (ni.sibling != NODE_NONE)
			ni.sibling = remap[ni.sibling];

		nb[o] = ni;
		o++;
	}
	
	check(o, ==, t->node_active);

	fprintf(stderr, "%zd -> %zd (total %zd)\n",
		t->node_top, t->node_active, t->node_total);
	
	t->node_top = o;
}

// *********************************************************************
static node_index tree_snapshot(tree *t)
{
	return t->node_active;
}

static void tree_rewind(tree *t, node_index snapshot, node_index root)
{
	t->node_active = snapshot;
	t->node_top = root;
}

// *********************************************************************
static void tree_write_include(tree *t, FILE *f)
{
	fprintf(f, "#define TREE_ROOT %zd\n", t->root);
	fprintf(f, "#define TREE_SIZE %zd\n", t->node_top);
	
	node *nb = t->node_buffer;
	node_index node_top = t->node_top;
	
	fprintf(f,"static node tree_nodes[TREE_SIZE] = {\n");
	
	for (node_index i = 0; i < node_top; i++) {
		fprintf(f, "{ %u, %u, %u, %zd, %zd }, // %zd %s\n",
			nb[i].hash, nb[i].depth,
			nb[i].guess, nb[i].child, nb[i].sibling,
			i, word_str(t->dictionary.words[nb[i].guess]).s);
	}
	
	fprintf(f, "};\n");
}

// ****************************************************************************
// 
// ****************************************************************************
#else
static node_index tree_push(tree *t, dict_index guess, depth_index depth)
{
	(void)guess;
	(void)depth;
	t->node_total++;
	return 0;
}

/*
static void tree_dummy(tree *t, dict_index guess)
{
	(void)guess;
	t->node_total++;
}
*/

static void tree_adopt(tree *t, node_index parent, node_index child,
	hash_index hash)
{
	(void)t;
	(void)parent;
	(void)child;
	(void)hash;
}

static void tree_prune(tree *t, node_index i)
{
	(void)t;
	(void)i;
}

static void tree_gc(tree *t, int force)
{
	(void)t;
	(void)force;
}

static void tree_write_include(tree *t, FILE *f)
{
	(void)t;
	(void)f;
}

#endif

// ****************************************************************************
// 
// ****************************************************************************
/*
  tree_dfs() is given a set of potential targets ('couldbe'),
  and determines a decision tree to disambiguate between them.
  The decision tree is optimal in the sense that it has minimum
  depth (however, a smaller tree of the same depth may exist).
  
  For every guess, winning is always an option, so [WIN] leaf nodes
  are not explored or stored.
  
  tree_dfs() is called only with couldbe_size >= 3.
  The returned tree depth is always at least depth + 1. Indeed,
  even with couldbe_size = 2, the optimal trees reach at least
  depth + 1. The only shallower tree is for couldbe_size = 1.
  
  depth + 0:     {A}          { A, B }                { A, B }
                  |              |                       |
               guess A        guess A                 guess C
                  |          /       \               /       \
  depth + 1:    [WIN]     [WIN]      { B }         { A }   { B }
                                       |             |       |
                                    guess B       guess A guess B
                                       |             |       |
                                     [WIN]         [WIN]   [WIN]
  
  tree_dfs() can be called with at most depth = DEPTH_MAX - 1.
  Therefore, we should initially have depth_max = DEPTH_MAX:
  
  DEPTH_MAX - 1:             { A, B, ... }
                                   |
                                guess X
                       /       /       \       \
  DEPTH_MAX:        { A }    { B }   { C }    { D }
                      |        |       |        |
                   guess A  guess B guess C  guess D
                      |        |       |        |
                    [WIN]    [WIN]   [WIN]    [WIN]
*/
typedef struct tree_tries_r_t {
	node_index best_node;
	depth_index best_depth;
} tree_tries_r;

static tree_tries_r tree_dfs(tree *t,
	depth_index depth, depth_index depth_ub,
	dict_index couldbe_size, dict_index *couldbe_set,
	dict_index guesses_size, dict_index *guesses_list)
{
	int hard = 0;
	word prev_g;
	hash_index prev_h;
	
	if ((t->flags & GAME_HARD) && (depth > 0)) {
		hard = 1;
		prev_g = t->dictionary.words[t->stack[depth - 1].v.g];
		prev_h = t->stack[depth - 1].v.h;
	}
	
	local v;
	
	// local variables and parameters
	memset(&v, 0, sizeof(local));
	
	v.depth_ub = depth_ub;
	v.couldbe_size = couldbe_size;
	v.guesses_size = guesses_size;
	
	// local variables
	level *l = &t->stack[depth];
	
#if 0
	if (hard) {
		fprintf(stderr, "depth %d:\n", depth);
		fprintf(stderr, "couldbe: %d\n", couldbe_size);
		if (couldbe_size < 100) {
			for (dict_index i = 0; i < couldbe_size; i++) {
				dict_index g = couldbe_set[i];
				
				fprintf(stderr, "%s\n",
					word_str(t->dictionary.words[g]).s);
			}
		}
	
		for (dict_index i = 0; i < couldbe_size; i++) {
			dict_index g = couldbe_set[i];
			word gw = t->dictionary.words[g];
			if (!hard_allowed(&t->table, prev_g, prev_h, gw)) {
				fprintf(stderr, "depth %d, "
					"mask masks a couldbe\n",
					depth);
				fprintf(stderr, "%s\n",
					word_str(t->dictionary.words[g]).s);
				check(0, ==, 1);
			}
		}
	}
#endif
	
	// maximum 'couldbe' cardinality for an answer:
	// an answer of cardinality couldbe_size yields
	// no refinement, which does not happen in an optimal tree
	dict_index answer_couldbe_ub = v.couldbe_size - 1;
	
	if (depth + 2 > v.depth_ub) {
		// each answer must have at most 1 target in its bucket
		answer_couldbe_ub = 1;
		
		// if we have more than DISAMBIGUATION (= 212) targets,
		// no guess can disambiguate them all; as a heuristic,
		// this value can be lowered (fast_c).
		if (v.couldbe_size > t->fast_c) {
			tree_tries_r r;
			
			r.best_depth = DEPTH_INF;
			r.best_node = NODE_NONE;
			return r;
		}

		// if 4 letters are identical in all targets,
		// at most 6 targets can be disambiguated
		// (because we guess 5 letters)
	}
	
	// XXX remove this later
	dict_index noinfo = 0;
	size_t avgj = 0;
	
	// sweep tries, determine best
	hash_index (*hash_cache)[DICT_SIZE] = l->hash_cache;
	dict_index *answer_couldbe_size = l->answer_couldbe_size;
	uint64_t *guess_score = l->guess_score;
	dict_index *guesses_sorted = l->guesses_sorted;
	dict_index *guesses_aux = l->guesses_aux;
	
	v.guesses_valid = 0;

	for (dict_index u = 0; u < v.guesses_size; u++) {
		dict_index g = (v.guesses_size == DICT_SIZE)
			? u : guesses_list[u];
		
		if (hard) {
			word gw = t->dictionary.words[g];
			if (!hard_allowed(&t->table, prev_g, prev_h, gw))
				continue;
		}
		
		hash_index *hash_cache_g = hash_cache[g];
		hash_index *lut_g = lut_lookup_guess(&t->table, g);
		
		// clear buckets
		for (hash_index h = 0; h < THREE_TO_LENGTH; h++)
			answer_couldbe_size[h] = 0;
	
		// count buckets
		int too_large = 0;
		
		for (dict_index j = 0; j < v.couldbe_size; j++) {
			dict_index x = couldbe_set[j];
			
			hash_index h = lut_g[x];
			hash_cache_g[j] = h;
			
			dict_index sz = ++answer_couldbe_size[h];
			
			if (sz > answer_couldbe_ub) {
				too_large = 1;
				avgj += j;
				break;
			}
		}
		
		if (too_large) {
			noinfo++;
			continue;
		}
		
		avgj += v.couldbe_size;
		
		uint64_t score = 0;
		
		for (hash_index h = 0; h < THREE_TO_LENGTH; h++) {
			uint64_t sz = answer_couldbe_size[h];
			
			score += sz * sz;
		}
		
		guess_score[g] = score;
		guesses_sorted[v.guesses_valid++] = g;
	}
	
	t->prog_ops += v.guesses_size * (THREE_TO_LENGTH + v.couldbe_size);
	
	aux_sort(v.guesses_valid, &guesses_sorted, &guesses_aux, guess_score);
	
	v.ord_cutoff = v.guesses_valid;
	if (v.ord_cutoff > t->fast_k)
		v.ord_cutoff = t->fast_k;

	// multithreaded exploration: divide root node
	dict_index ord0 = 0;
	dict_index ord1 = v.ord_cutoff;
	
	if (depth == 0) {
		ord0 = (size_t)v.ord_cutoff * t->thread_id / t->thread_total;
		ord1 = (size_t)v.ord_cutoff * (t->thread_id + 1)
			/ t->thread_total;
		if ((ord0 != 0) || (ord1 != v.ord_cutoff)) {
			fprintf(stderr, "Thread %zd / %zd: tries %d to %d.\n",
				t->thread_id, t->thread_total, ord0, ord1 - 1);
		}
	}
	
	/*if ((t->prog_ops % 500) == 0) {
		fprintf(stderr,
			"d %2d: noinfo %5d / %5d, cb = %5d, avg j = %.2f"
			" => ord_cutoff %d / %d\n",
			depth, noinfo, v.guesses_size, couldbe_size,
			(double)avgj / v.guesses_size,
			v.ord_cutoff, v.guesses_valid);
		for (dict_index q = 0; q < v.couldbe_size; q++) {
			dict_index x = couldbe_set[q];
			fprintf(stderr, " %s",
				word_str(t->dictionary.words[x]).s);
		}
		fprintf(stderr, "\n");
	}// */
	
	// sweep tries, recurse
	v.best_depth = DEPTH_INF;
	v.best_node = NODE_NONE;
	
	hash_index *answer_list = l->answer_list;
	dict_index *answer_couldbe_offs = l->answer_couldbe_offs;
	dict_index *answer_couldbe_set = l->answer_couldbe_set;
	
	for (v.ord = ord0; v.ord < ord1; v.ord++) {
		v.g = guesses_sorted[v.ord];

		// clear buckets
		v.answer_count = 0;
		
		for (hash_index h = 0; h < THREE_TO_LENGTH; h++)
			answer_couldbe_size[h] = 0;
	
		// re-determine bucket sizes
		hash_index *hash_cache_g = hash_cache[v.g];
		
		for (dict_index j = 0; j < v.couldbe_size; j++) {
			hash_index h = hash_cache_g[j];
			
			dict_index sz = answer_couldbe_size[h];
			if (sz == 0) {
				answer_list[v.answer_count] = h;
				v.answer_count++;
			}
			
			answer_couldbe_size[h] = sz + 1;
		}
		
		// compute offsets
		dict_index offs = 0;
		for (hash_index a = 0; a < v.answer_count; a++) {
			hash_index h = answer_list[a];
			answer_couldbe_offs[h] = offs;
			offs += answer_couldbe_size[h];
		}
		
		check(offs, ==, v.couldbe_size);
		
		// fill buckets
		for (dict_index j = 0; j < v.couldbe_size; j++) {
			hash_index h = hash_cache_g[j];
			dict_index o = answer_couldbe_offs[h]++;
			answer_couldbe_set[o] = couldbe_set[j];
		}

		// sweep buckets
		l->v = v;
		tree_gc(t, 0);
		v = l->v;
		
		node_index snapshot = tree_snapshot(t);
		v.cur_node = tree_push(t, v.g, depth);
		v.depth_max = 0;
		
		//tree_progress(t, depth, " PRE");
		//fprintf(stderr, "Loop begin %d\n", depth);
		for (v.a = 0; v.a < v.answer_count; v.a++) {
			v.h = answer_list[v.a];
			
			// note: by now, offsets are shifted by size
			v.asize = answer_couldbe_size[v.h];
			dict_index *aset = answer_couldbe_set
				+ answer_couldbe_offs[v.h]
				- v.asize;
			
			tree_tries_r sub_r;
			
			if (v.asize == 1) {
				if (v.h == HASH_ALL_MATCH)
					continue;
				
				// singleton answer, no recursion
				sub_r.best_depth = depth + 1;
				sub_r.best_node =
					tree_push(t, aset[0], depth + 1);
				//fprintf(stderr, "Push1 %d\n", depth);
			} else if (depth + 2 > v.depth_ub) {
				// need recursion, subtree_depth >= depth + 2
				v.depth_max = DEPTH_INF;
				//fprintf(stderr, "BREAK %d: %d > %d\n",
				//	depth, depth + 2, depth_ub);
				break;
			} else if (v.asize == 2) {
				// two-targets answer, no recursion
				// push nodes for optimal two-target tree
				sub_r.best_depth = depth + 2;
				sub_r.best_node =
					tree_push(t, aset[0], depth + 1);
				tree_adopt(t, sub_r.best_node,
					tree_push(t, aset[1], depth + 2),
					HASH_ANY);
				//fprintf(stderr, "Push2 %d\n", depth);
			} else {
				// recurse
				/*fprintf(stderr, "Recurse depth %d\n"
					"     guess %s hash %d\n",
					depth,
					word_str(t->dictionary.words[guess]).s,
					h);
				
				lut_print_hash(stderr, h);
				*/
				l->v = v;
				
				sub_r = tree_dfs(t,
					depth + 1, v.depth_ub,
					v.asize, aset,
					v.guesses_valid,
					guesses_sorted);
				
				v = l->v;
				
				//fprintf(stderr, "End recurse depth %d, r %d\n",
				//	depth, sub_r.best_depth);
				
				if (sub_r.best_depth == DEPTH_INF) {
					v.depth_max = DEPTH_INF;
					break;
				}
			}
			
			// adopt subtree
			tree_adopt(t, v.cur_node, sub_r.best_node, v.h);

			// update max
			if (v.depth_max < sub_r.best_depth) {
				v.depth_max = sub_r.best_depth;
				
				if (v.depth_max >= v.best_depth)
					break;
			}
		}

		/*fprintf(stderr, "Loop ends %d\n", depth);
		tree_progress(t, depth, " P2");
		if (v.depth_ub >= DEPTH_MAX) {
			check_s(l->max_depth, <, DEPTH_INF,
				"depth = %d\n", depth);
		}*/
		
		// update best guess
		if (v.best_depth > v.depth_max) {
			if (v.best_node != NODE_NONE)
				tree_prune(t, v.best_node);
			
			v.best_depth = v.depth_max;
			v.best_node = v.cur_node;
			check(v.depth_ub, >, v.depth_max - 1);
			v.depth_ub = v.depth_max - 1;
			
			// target disambiguated
			if (v.depth_max <= depth + 1)
				break;
		} else {
			tree_rewind(t, snapshot, v.cur_node);
		}

		l->v = v;
		tree_progress(t, depth, "");
	}
	
	if (depth == 0) {
		l->v = v;
		tree_gc(t, 1);
		v = l->v;
	}
	
	tree_tries_r r;
	
	r.best_depth = v.best_depth;
	r.best_node = v.best_node;
	
	l->v.cur_node = NODE_NONE;
	l->v.best_node = NODE_NONE;
	/*if (v.depth_ub >= DEPTH_MAX) {
		check_s(r.best_depth, <, DEPTH_INF,
			"at depth %d, depth_ub = %d = %d, "
			"cb = %d, ord1 = %d %d\n",
			depth, depth_ub, v.depth_ub, couldbe_size,
			ord1, ord_cutoff);
	}*/
	return r;
}

// ****************************************************************************
// 
// ****************************************************************************
tree *tree_build(size_t thread_id, size_t thread_total,
	depth_index global_depth_ub,
	dict_index fast_k, dict_index fast_c, size_t prog_i, int flags)
{
	// parameters and safety checks
	fprintf(stderr, "Game mode:%s%s%s\n",
		((flags & (GAME_HARD | GAME_CHEAT)) == 0) ? " regular" : "",
		(flags & GAME_HARD) ? " hard" : "",
		(flags & GAME_CHEAT) ? " cheat" : "");
	
	fprintf(stderr, "Depth: %d\n", global_depth_ub);
	
	int exact = 1;
	
	if ((fast_k == 0) || (fast_k >= DICT_SIZE))
		fast_k = DICT_SIZE;
	else
		exact = 0;
	
	if ((fast_c == 0) || (fast_c >= DISAMBIGUATION))
		fast_c = DISAMBIGUATION;
	else
		exact = 0;

	if (exact)
		fprintf(stderr, "Exact tree search\n");
	else
		fprintf(stderr, "Heuristic tree search, K = %d, C = %d\n",
			fast_k, fast_c);
	
	if (prog_i == 0) {
		prog_i = (size_t)4 << 30;
	}

	size_t tsize = sizeof(tree) + (size_t)global_depth_ub * sizeof(level);
	
	fprintf(stderr, "Allocating tree (%.1f GB)...\n",
		(double)tsize / ((size_t)1 << 30));
	
	tree *t = malloc(tsize);
	
	if (!t) {
		fprintf(stderr, "Tree allocation failed (%zd bytes)\n", tsize);
		exit(1);
		
	}
	
	memset(t, 0, tsize);
	
	t->flags = flags;
	t->start_time = 0;
	t->fast_k = fast_k;
	t->fast_c = fast_c;
	t->prog_i = prog_i;
	t->prog_ops = 0;
	t->prog_last = 0;
	t->prog_next = 0;
	
	t->global_depth_ub = global_depth_ub;
	
	// builtin dictionary
	fprintf(stderr, "Loading dictionary...\n");
	dict_init(&t->dictionary);
	dict_builtin_full(&t->dictionary);
	
	dict_index couldbe_size;

	if (flags & GAME_CHEAT) {
		fprintf(stderr, "Building cheat mode dictionary...\n");
		dict sol;
		
		dict_init(&sol);
		dict_builtin_solutions(&sol);
		
		couldbe_size = 0;
		
		for (dict_index i = 0; i < DICT_SIZE; i++) {
			word w = t->dictionary.words[i];
			
			if (dict_has(&sol, w))
				t->couldbe[couldbe_size++] = i;
		}
		
		dict_clear(&sol);
	} else {
		couldbe_size = DICT_SIZE;
	
		for (dict_index i = 0; i < DICT_SIZE; i++)
			t->couldbe[i] = i;
	}

	fprintf(stderr, "Building lookup table...\n");
	lut_build(&t->table, &t->dictionary);
	
	fprintf(stderr, "Building tree...\n");

	for (depth_index d = 0; d < t->global_depth_ub; d++)
		t->stack[d].v.cur_node = NODE_NONE;

	t->start_time = time(NULL);
	t->prog_time = t->start_time;
	t->root = NODE_NONE;
	t->node_top = t->node_active = t->node_total = 0;
	t->thread_id = thread_id;
	t->thread_total = thread_total;

	tree_tries_r root = tree_dfs(t, 0, t->global_depth_ub,
		couldbe_size, t->couldbe, DICT_SIZE, NULL);
	
	time_t final_delta = time(NULL) - t->start_time;

	fprintf(stderr, "# Finished: %5zds, op: %6zd G, "
		"op/s: %5zd M\n",
		final_delta, t->prog_ops >> 30,
		(t->prog_ops / final_delta) >> 20);
	
	if (root.best_depth == DEPTH_INF) {
		fprintf(stderr, "Done. No tree found.\n");
	} else {
		fprintf(stderr, "Done. Best tree: nodes = %zd, depth = %d.\n",
			t->node_active, root.best_depth);
	}
	
	t->root = root.best_node;
	
	if (root.best_depth < DEPTH_INF)
		tree_write_include(t, stdout);
	
	return t;
}

// ****************************************************************************
// 
// ****************************************************************************
void tree_get(tree *t,
	node_index *count_r, node_index *root_r, node **nodes_r)
{
	if (count_r)
		*count_r = t->node_active;
	if (root_r)
		*root_r = t->root;
	if (nodes_r)
		*nodes_r = t->node_buffer;
}

// ****************************************************************************
// 
// ****************************************************************************
void tree_free(tree *t)
{
	dict_clear(&t->dictionary);
	free(t);
}

// ****************************************************************************
// 
// ****************************************************************************
